from typing import Optional
import docker
from llm_sandbox.docker import SandboxDockerSession, ConsoleOutput
from llm_sandbox.const import SupportedLanguage
from docker.types import Mount


class MicromambaSession(SandboxDockerSession):
    """
    SandboxDockerSession does not allow activation of micromamba environment,
    this class extends it and allows that which makes it possible for LLM agents to installed conda dependencies.
    """

    def __init__(
        self,
        client: Optional[docker.DockerClient] = None,
        image: Optional[str] = "mambaorg/micromamba:latest",
        dockerfile: Optional[str] = None,
        lang: str = SupportedLanguage.PYTHON,
        keep_template: bool = False,
        verbose: bool = False,
        mounts: Optional[list[Mount]] = None,
        environment: str = "base",
    ):
        super().__init__(
            client=client,
            image=image,
            dockerfile=dockerfile,
            lang=lang,
            keep_template=keep_template,
            verbose=verbose,
            mounts=mounts,
        )
        self.environment = environment

    def execute_command(
        self, command: Optional[str], workdir: Optional[str] = None
    ) -> ConsoleOutput:
        if not command:
            raise ValueError("Command cannot be empty")

        if not self.container:
            raise RuntimeError(
                "Session is not open. Please call open() method before executing commands."
            )
        command = f"micromamba run -n {self.environment} {command}"

        if self.verbose:
            print(f"Executing command: {command}")

        if workdir:
            exit_code, exec_log = self.container.exec_run(
                command, stream=True, tty=True, workdir=workdir
            )
        else:
            exit_code, exec_log = self.container.exec_run(
                command, stream=True, tty=True
            )

        output = ""
        if self.verbose:
            print("Output:", end=" ")

        for chunk in exec_log:
            chunk_str = chunk.decode("utf-8")
            output += chunk_str
            if self.verbose:
                print(chunk_str, end="")

        return ConsoleOutput(text=output, exit_code=exit_code)



"""Main session module for LLM Sandbox."""

from typing import Optional, TypeVar, cast, TYPE_CHECKING
from importlib.util import find_spec

from .const import SupportedLanguage, SandboxBackend
from .factory import UnifiedSessionFactory
from .exceptions import DependencyError
from .base import Session

# Type checking imports that won't be evaluated at runtime
if TYPE_CHECKING:
    pass

T = TypeVar("T", bound=Session)


def _check_dependency(backend: SandboxBackend) -> None:
    """Check if required dependency is installed for the given backend."""
    if backend == SandboxBackend.DOCKER or backend == SandboxBackend.MICROMAMBA:
        if not find_spec("docker"):
            raise DependencyError(
                "Docker backend requires 'docker' package. Install it with: pip install llm-sandbox[docker]"
            )
    elif backend == SandboxBackend.KUBERNETES:
        if not find_spec("kubernetes"):
            raise DependencyError(
                "Kubernetes backend requires 'kubernetes' package. Install it with: pip install llm-sandbox[k8s]"
            )
    elif backend == SandboxBackend.PODMAN:
        if not find_spec("podman"):
            raise DependencyError(
                "Podman backend requires 'podman' package. Install it with: pip install llm-sandbox[podman]"
            )


class ContextManagerMixin:
    """Mixin to add context manager support to session instances."""

    def __enter__(self):
        """Enter the context manager."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context manager."""
        if hasattr(self, "close"):
            self.close()


class SandboxSession:
    """Factory class for creating sandbox sessions with context manager support."""

    def __new__(
        cls,
        backend: SandboxBackend = SandboxBackend.DOCKER,
        image: Optional[str] = None,
        dockerfile: Optional[str] = None,
        lang: str = SupportedLanguage.PYTHON,
        keep_template: bool = False,
        commit_container: bool = True,
        verbose: bool = False,
        runtime_configs: Optional[dict] = None,
        **kwargs,
    ) -> "SandboxSession":
        """
        Create a new sandbox session.

        Args:
            backend: Docker, Kubernetes or Podman backend
            image: Container image to use
            dockerfile: Path to Dockerfile
            lang: Programming language
            keep_template: Whether to keep the container template
            commit_container: Whether to commit container changes
            verbose: Enable verbose logging
            runtime_configs: Additional runtime configurations, check the specific backend for more details

        Returns:
            A new sandbox session of the appropriate type (Docker, Kubernetes, Podman, or Micromamba)

        Raises:
            DependencyError: If the required dependency for the chosen backend is not installed
        """
        # Check if required dependency is installed
        _check_dependency(backend)

        factory = UnifiedSessionFactory()

        session = factory.create_session(
            backend=backend,
            image=image,
            dockerfile=dockerfile,
            lang=lang,
            keep_template=keep_template,
            commit_container=commit_container,
            verbose=verbose,
            runtime_configs=runtime_configs,
            **kwargs,
        )

        # Create a new class that inherits from both the session class and our mixin
        session.__class__ = type(
            "SandboxSession",
            (session.__class__, ContextManagerMixin, SandboxSession),
            {},
        )

        return cast("SandboxSession", session)



"""Factory for creating sandbox sessions."""

from typing import Union, Type, TYPE_CHECKING
from abc import ABC, abstractmethod

from .const import SandboxBackend
from .exceptions import ValidationError

if TYPE_CHECKING:
    from .docker import SandboxDockerSession
    from .kubernetes import SandboxKubernetesSession
    from .podman import SandboxPodmanSession
    from .micromamba import MicromambaSession


class SessionFactory(ABC):
    """Abstract base factory for creating sandbox sessions."""

    @abstractmethod
    def create_session(
        self, **kwargs
    ) -> Union[
        "SandboxDockerSession",
        "SandboxKubernetesSession",
        "SandboxPodmanSession",
        "MicromambaSession",
    ]:
        """Create a new sandbox session."""
        raise NotImplementedError


class DockerSessionFactory(SessionFactory):
    """Factory for creating Docker-based sandbox sessions."""

    def create_session(self, **kwargs) -> "SandboxDockerSession":
        """Create a new Docker sandbox session."""
        from .docker import SandboxDockerSession

        return SandboxDockerSession(**kwargs)


class KubernetesSessionFactory(SessionFactory):
    """Factory for creating Kubernetes-based sandbox sessions."""

    def create_session(self, **kwargs) -> "SandboxKubernetesSession":
        """Create a new Kubernetes sandbox session."""
        from .kubernetes import SandboxKubernetesSession

        return SandboxKubernetesSession(**kwargs)


class PodmanSessionFactory(SessionFactory):
    """Factory for creating Podman-based sandbox sessions."""

    def create_session(self, **kwargs) -> "SandboxPodmanSession":
        """Create a new Podman sandbox session."""
        from .podman import SandboxPodmanSession

        return SandboxPodmanSession(**kwargs)


class MicromambaSessionFactory(SessionFactory):
    """Factory for creating Micromamba-based sandbox sessions."""

    def create_session(self, **kwargs) -> "MicromambaSession":
        """Create a new Micromamba sandbox session."""
        from .micromamba import MicromambaSession

        return MicromambaSession(**kwargs)


class UnifiedSessionFactory:
    """Unified factory for creating sandbox sessions of any type."""

    def create_session(
        self, backend: SandboxBackend = SandboxBackend.DOCKER, **kwargs
    ) -> Union[
        "SandboxDockerSession",
        "SandboxKubernetesSession",
        "SandboxPodmanSession",
        "MicromambaSession",
    ]:
        """Create a new sandbox session of the specified type."""
        factories: dict[SandboxBackend, Type[SessionFactory]] = {
            SandboxBackend.DOCKER: DockerSessionFactory,
            SandboxBackend.KUBERNETES: KubernetesSessionFactory,
            SandboxBackend.PODMAN: PodmanSessionFactory,
            SandboxBackend.MICROMAMBA: MicromambaSessionFactory,
        }

        factory_class = factories.get(backend)
        if factory_class is None:
            raise ValidationError(f"Unsupported backend: {backend}")

        return factory_class().create_session(**kwargs)



"""Custom exceptions for LLM Sandbox."""


class SandboxException(Exception):
    """Base exception for all sandbox related errors."""

    pass


class ContainerError(SandboxException):
    """Raised when container operations fail."""

    pass


class SecurityError(SandboxException):
    """Raised for security related issues."""

    pass


class ResourceError(SandboxException):
    """Raised when resource limits are exceeded."""

    pass


class TimeoutError(SandboxException):
    """Raised when execution exceeds timeout limit."""

    pass


class ValidationError(SandboxException):
    """Raised when input validation fails."""

    pass


class DependencyError(SandboxException):
    """Raised when a required dependency is not installed."""

    pass



import io
import os
import tarfile
import tempfile
from typing import List, Optional, Union
from collections.abc import Iterator

from podman import PodmanClient
from podman.errors import NotFound
from podman.domain.containers import Container
from podman.domain.images import Image
from llm_sandbox.utils import (
    get_libraries_installation_command,
    get_code_file_extension,
    get_code_execution_command,
)
from llm_sandbox.base import Session, ConsoleOutput
from llm_sandbox.const import (
    SupportedLanguage,
    SupportedLanguageValues,
    DefaultImage,
    NotSupportedLibraryInstallation,
)


class SandboxPodmanSession(Session):
    def __init__(
        self,
        client: Optional[PodmanClient] = None,
        image: Optional[str] = None,
        dockerfile: Optional[str] = None,
        lang: str = SupportedLanguage.PYTHON,
        keep_template: bool = False,
        commit_container: bool = True,
        verbose: bool = False,
        mounts: Optional[list] = None,
        runtime_configs: Optional[dict] = None,
        **kwargs,
    ):
        """
        Create a new sandbox session
        :param client: Podman client, if not provided, a new client will be created based on local podman context
        :param image: Podman image to use
        :param dockerfile: Path to the Dockerfile, if image is not provided
        :param lang: Language of the code
        :param keep_template: if True, the image and container will not be removed after the session ends
        :param commit_container: if True, the podman container will be commited after the session ends
        :param verbose: if True, print messages
        :param mounts: List of mounts to be mounted to the container
        :param runtime_configs: Additional configurations for the container, i.e. resources limits (cpu_count, mem_limit), etc.
        """
        super().__init__(lang, verbose)
        if image and dockerfile:
            raise ValueError("Only one of image or dockerfile should be provided")

        if lang not in SupportedLanguageValues:
            raise ValueError(
                f"Language {lang} is not supported. Must be one of {SupportedLanguageValues}"
            )

        if not image and not dockerfile:
            image = DefaultImage.__dict__[lang.upper()]

        self.lang: str = lang
        self.client: Optional[PodmanClient] = client or PodmanClient()
        self.image: Union[Image, str] = image
        self.dockerfile: Optional[str] = dockerfile
        self.container: Optional[Container] = None
        self.path = None
        self.keep_template = keep_template
        self.commit_container = commit_container
        self.is_create_template: bool = False
        self.verbose = verbose
        self.mounts = mounts
        self.runtime_configs = runtime_configs

    def open(self):
        warning_str = (
            "Since the `keep_template` flag is set to True, the Podman image will not be removed after the session ends "
            "and remains for future use."
        )

        # Build image if a Dockerfile is provided
        if self.dockerfile:
            self.path = os.path.dirname(self.dockerfile)
            if self.verbose:
                f_str = f"Building Podman image from {self.dockerfile}"
                f_str = f"{f_str}\n{warning_str}" if self.keep_template else f_str
                print(f_str)

            self.image, _ = self.client.images.build(
                path=self.path,
                dockerfile=os.path.basename(self.dockerfile),
                tag=f"sandbox-{self.lang.lower()}-{os.path.basename(self.path)}",
            )
            self.is_create_template = True

        # Check or pull the image
        if isinstance(self.image, str):
            try:
                # Try to get the image locally
                self.image = self.client.images.get(self.image)
                if self.verbose:
                    print(f"Using image {self.image.tags[-1]}")
            except NotFound:
                if self.verbose:
                    print(
                        f"Image {self.image} not found locally. Attempting to pull..."
                    )

                try:
                    # Attempt to pull the image
                    self.image = self.client.images.pull(self.image)
                    if self.verbose:
                        print(f"Successfully pulled image {self.image.tags[-1]}")
                    self.is_create_template = True
                except Exception as e:
                    raise RuntimeError(f"Failed to pull image {self.image}: {e}")

        # Ensure mounts is an iterable (empty list if None)
        mounts = self.mounts if self.mounts is not None else []

        # Create the container
        self.container = self.client.containers.create(
            image=self.image.id if isinstance(self.image, Image) else self.image,
            tty=True,
            mounts=mounts,  # Use the adjusted mounts
            **self.runtime_configs if self.runtime_configs else {},
        )
        self.container.start()

    def close(self):
        if self.container:
            if self.commit_container and isinstance(self.image, Image):
                # Extract repository and tag
                if self.image.tags:
                    full_tag = self.image.tags[-1]
                    if ":" in full_tag:
                        repository, tag = full_tag.rsplit(":", 1)
                    else:
                        repository = full_tag
                        tag = "latest"
                try:
                    # Commit the container with repository and tag
                    self.container.commit(repository=repository, tag=tag)
                    if self.verbose:
                        print(f"Committed container as image {repository}:{tag}")
                except Exception as e:
                    if self.verbose:
                        print(f"Failed to commit container: {e}")
                    raise

            # Stop and remove the container
            self.container.stop()
            self.container.remove(force=True)
            self.container = None

        if self.is_create_template and not self.keep_template:
            # check if the image is used by any other container
            containers = self.client.containers.list(all=True)
            image_id = (
                self.image.id
                if isinstance(self.image, Image)
                else self.client.images.get(self.image).id
            )
            image_in_use = any(
                container.image.id == image_id for container in containers
            )

            if not image_in_use:
                if isinstance(self.image, str):
                    self.client.images.remove(self.image)
                elif isinstance(self.image, Image):
                    self.image.remove(force=True)
                else:
                    raise ValueError("Invalid image type")
            else:
                if self.verbose:
                    print(
                        f"Image {self.image.tags[-1]} is in use by other containers. Skipping removal.."
                    )

    def run(self, code: str, libraries: Optional[List] = None) -> ConsoleOutput:
        if not self.container:
            raise RuntimeError(
                "Session is not open. Please call open() method before running code."
            )

        if libraries:
            if self.lang.upper() in NotSupportedLibraryInstallation:
                raise ValueError(
                    f"Library installation has not been supported for {self.lang} yet!"
                )
            if self.lang == SupportedLanguage.GO:
                self.execute_command("mkdir -p /example")
                self.execute_command("go mod init example", workdir="/example")
                self.execute_command("go mod tidy", workdir="/example")

                for library in libraries:
                    command = get_libraries_installation_command(self.lang, library)
                    _ = self.execute_command(command, workdir="/example")
            else:
                for library in libraries:
                    command = get_libraries_installation_command(self.lang, library)
                    _ = self.execute_command(command)
        with tempfile.TemporaryDirectory() as directory_name:
            code_file = os.path.join(
                directory_name, f"code.{get_code_file_extension(self.lang)}"
            )
            if self.lang == SupportedLanguage.GO:
                code_dest_file = "/example/code.go"
            else:
                code_dest_file = (
                    f"/tmp/code.{get_code_file_extension(self.lang)}"  # code_file
                )

            with open(code_file, "w") as f:
                f.write(code)

            self.copy_to_runtime(code_file, code_dest_file)

            output = ConsoleOutput(exit_code=0, text="")
            commands = get_code_execution_command(self.lang, code_dest_file)
            for command in commands:
                if self.lang == SupportedLanguage.GO:
                    output = self.execute_command(command, workdir="/example")
                else:
                    output = self.execute_command(command)

            return output

    def copy_from_runtime(self, src: str, dest: str):
        if not self.container:
            raise RuntimeError(
                "Session is not open. Please call open() method before copying files."
            )

        if self.verbose:
            print(f"Copying {self.container.short_id}:{src} to {dest}..")

        bits, stat = self.container.get_archive(src)
        if stat["size"] == 0:
            raise FileNotFoundError(f"File {src} not found in the container")

        tarstream = io.BytesIO(b"".join(bits))
        with tarfile.open(fileobj=tarstream, mode="r") as tar:
            tar.extractall(os.path.dirname(dest))

    def copy_to_runtime(self, src: str, dest: str):
        if not self.container:
            raise RuntimeError(
                "Session is not open. Please call open() method before copying files."
            )

        is_created_dir = False
        directory = os.path.dirname(dest)
        if directory and not self.container.exec_run(f"test -d {directory}")[0] == 0:
            self.container.exec_run(f"mkdir -p {directory}")
            is_created_dir = True

        if self.verbose:
            if is_created_dir:
                print(f"Creating directory {self.container.short_id}:{directory}")
            print(f"Copying {src} to {self.container.short_id}:{dest}..")

        tarstream = io.BytesIO()
        with tarfile.open(fileobj=tarstream, mode="w") as tar:
            tar.add(src, arcname=os.path.basename(src))

        tarstream.seek(0)
        self.container.put_archive(os.path.dirname(dest), tarstream)

    def execute_command(
        self, command: Optional[str], workdir: Optional[str] = None
    ) -> ConsoleOutput:
        if not command:
            raise ValueError("Command cannot be empty")

        if not self.container:
            raise RuntimeError(
                "Session is not open. Please call open() method before executing commands."
            )

        if self.verbose:
            print(f"Executing command: {command}")

        if workdir:
            exit_code, exec_log = self.container.exec_run(
                command, stream=True, tty=True, workdir=workdir
            )
        else:
            exit_code, exec_log = self.container.exec_run(
                command, stream=True, tty=True
            )
        if isinstance(exec_log, Iterator):
            output = ""
            for chunk in exec_log:
                chunk_str = chunk.decode("utf-8")
                output += chunk_str
                if self.verbose:
                    print(chunk_str, end="")
        else:
            output = exec_log.decode("utf-8")

        if self.verbose:
            print(output)

        return ConsoleOutput(text=output, exit_code=exit_code)



"""Base session functionality for LLM Sandbox."""

import logging
from abc import ABC, abstractmethod
from typing import Optional, Dict
from dataclasses import dataclass


@dataclass
class ExecutionResult:
    """Result of code execution in sandbox."""

    exit_code: int
    output: str
    error: str
    execution_time: float
    resource_usage: Dict


class ConsoleOutput:
    """Console output from code execution."""

    def __init__(self, text: str, exit_code: int = 0):
        self._text = text
        self._exit_code = exit_code

    @property
    def text(self) -> str:
        return self._text

    @property
    def exit_code(self) -> int:
        return self._exit_code

    def __repr__(self):
        return f"ConsoleOutput(text={self.text}, exit_code={self.exit_code})"

    def __str__(self):
        return self.text


class Session(ABC):
    """Abstract base class for sandbox sessions."""

    def __init__(
        self,
        lang: str,
        verbose: bool = True,
        strict_security: bool = True,
        runtime_configs: Optional[dict] = None,
        logger: Optional[logging.Logger] = None,
    ):
        self.lang = lang
        self.verbose = verbose
        self.runtime_configs = runtime_configs
        self.strict_security = strict_security
        self.logger = logger or logging.getLogger(__name__)

    def _log(self, message: str, level: str = "info"):
        """Log message if verbose is enabled."""
        if self.verbose:
            getattr(self.logger, level)(message)

    @abstractmethod
    def open(self):
        """Open the sandbox session."""
        raise NotImplementedError

    @abstractmethod
    def close(self):
        """Close the sandbox session."""
        raise NotImplementedError

    @abstractmethod
    def copy_to_runtime(self, src: str, dest: str):
        """Copy file to sandbox runtime."""
        raise NotImplementedError

    @abstractmethod
    def copy_from_runtime(self, src: str, dest: str):
        """Copy file from sandbox runtime."""
        raise NotImplementedError

    @abstractmethod
    def execute_command(self, command: str) -> ConsoleOutput:
        """Execute command in sandbox."""
        raise NotImplementedError

    def __enter__(self):
        """Context manager entry."""
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit."""
        self.close()



"""Security scanning functionality for LLM Sandbox."""

import re
from typing import List, Pattern, Dict
from dataclasses import dataclass
from .exceptions import SecurityError


@dataclass
class SecurityIssue:
    """Represents a security issue found in code."""

    pattern: str
    description: str
    severity: str  # 'high', 'medium', 'low'
    line_number: int


class SecurityScanner:
    """Scanner for detecting potential security issues in code."""

    def __init__(self):
        self.patterns: Dict[str, Pattern] = {  # type: ignore[annotation-unchecked]
            "system_calls": re.compile(r"os\.system|subprocess\."),
            "code_execution": re.compile(r"eval\(|exec\(|compile\("),
            "file_operations": re.compile(r"open\(|file\(|read|write"),
            "network_access": re.compile(r"socket\.|urllib|requests\."),
            "shell_injection": re.compile(r"shell=True|commands\."),
            "dangerous_imports": re.compile(r"import\s+(os|subprocess|sys|shutil)"),
        }

        self.severity_map = {
            "system_calls": "high",
            "code_execution": "high",
            "file_operations": "medium",
            "network_access": "medium",
            "shell_injection": "high",
            "dangerous_imports": "medium",
        }

        self.descriptions = {
            "system_calls": "Direct system command execution detected",
            "code_execution": "Dynamic code execution functionality detected",
            "file_operations": "File system operations detected",
            "network_access": "Network access functionality detected",
            "shell_injection": "Potential shell injection vulnerability",
            "dangerous_imports": "Import of potentially dangerous modules",
        }

    def scan_code(self, code: str, strict: bool = True) -> List[SecurityIssue]:
        """
        Scan code for security issues.

        Args:
            code: The code to scan
            strict: If True, raise SecurityError for high severity issues

        Returns:
            List of SecurityIssue objects

        Raises:
            SecurityError: If strict=True and high severity issues are found
        """
        issues = []
        lines = code.split("\n")

        for line_num, line in enumerate(lines, 1):
            for pattern_name, pattern in self.patterns.items():
                if pattern.search(line):
                    issue = SecurityIssue(
                        pattern=pattern_name,
                        description=self.descriptions[pattern_name],
                        severity=self.severity_map[pattern_name],
                        line_number=line_num,
                    )
                    issues.append(issue)

        if strict and any(issue.severity == "high" for issue in issues):
            high_severity_issues = [i for i in issues if i.severity == "high"]
            raise SecurityError(
                "High severity security issues found:\n"
                + "\n".join(
                    f"Line {i.line_number}: {i.description}"
                    for i in high_severity_issues
                )
            )

        return issues

    def is_safe(self, code: str) -> bool:
        """Quick check if code is safe (no security issues)."""
        try:
            return len(self.scan_code(code, strict=True)) == 0
        except SecurityError:
            return False



"""LLM Sandbox - A lightweight and portable LLM sandbox runtime."""

from .session import SandboxSession  # noqa: F401
from .const import SupportedLanguage, SandboxBackend  # noqa: F401
from .exceptions import (  # noqa: F401
    SandboxException,
    ContainerError,
    SecurityError,
    ResourceError,
    TimeoutError,
    ValidationError,
    DependencyError,
)
